# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinShare library

    @copyright: 2011, 2012, 2013 by Paul Boddie <paul@boddie.org.uk>
    @copyright: 2003-2006 Edgewall Software
    @copyright: 2006 MoinMoin:AlexanderSchremmer
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from ContentTypeSupport import getContentPreferences
from DateSupport import getCurrentTime, getDateTimeFromRFC2822
from MoinSupport import *
from ItemSupport import ItemStore
from MoinMessage import GPG, is_encrypted, is_signed, MoinMessageError
from MoinMessageSupport import get_homedir, get_username_for_fingerprint
from MoinMoin.support.htmlmarkup import HTMLParseError, HTMLSanitizer, Markup
from MoinMoin import wikiutil
from email.parser import Parser
from codecs import getwriter

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

_getFragments = getFragments

__version__ = "0.1"

# More Moin 1.9 compatibility functions.

def has_member(request, groupname, username):
    if hasattr(request.dicts, "has_member"):
        return request.dicts.has_member(groupname, username)
    else:
        return username in request.dicts.get(groupname, [])

# Fragments employ a "moinshare" attribute.

fragment_attribute = "moinshare"

def getFragments(s):

    "Return all fragments in 's' having the MoinShare fragment attribute."

    fragments = []
    for format, attributes, body in _getFragments(s):
        if attributes.has_key(fragment_attribute):
            fragments.append((format, attributes, body))
    return fragments

def getPreferredOutputTypes(request, mimetypes):

    """
    Using the 'request', perform content negotiation, obtaining mimetypes common
    to the fragment (given by 'mimetypes') and the client (found in the Accept
    header).
    """

    accept = getHeader(request, "Accept", "HTTP")
    if accept:
        prefs = getContentPreferences(accept)
        return prefs.get_preferred_types(mimetypes)
    else:
        return mimetypes

def getUpdatedTime(metadata):

    """
    Return the last updated time based on the given 'metadata', using the
    current time if no explicit last modified time is specified.
    """

    # NOTE: We could attempt to get the last edit time of a fragment.

    latest_timestamp = metadata.get("last-modified")
    if latest_timestamp:
        return latest_timestamp
    else:
        return getCurrentTime()

# Entry/update classes.

class Update:

    "A feed update entry."

    def __init__(self):
        self.title = None
        self.link = None
        self.content = None
        self.content_type = None
        self.updated = None
        self.author = None

        # Page-related attributes.

        self.fragment = None
        self.preferred = None

        # Message-related attributes.

        self.message_number = None
        self.parts = None

        # Message- and page-related attributes.

        self.page = None

        # Identification.

        self.path = []

    def unique_id(self):
        return "moinshare-tab-%s-%s" % (self.message_number, "-".join(map(str, self.path)))

    def __cmp__(self, other):
        if self.updated is None and other.updated is not None:
            return 1
        elif self.updated is not None and other.updated is None:
            return -1
        else:
            return cmp(self.updated, other.updated)

    def copy(self, part_number=None):
        update = Update()
        update.title = self.title
        update.link = self.link
        update.updated = self.updated
        update.author = self.author
        update.fragment = self.fragment
        update.preferred = self.preferred
        update.message_number = self.message_number
        update.page = self.page
        update.path = self.path[:]
        if part_number is not None:
            update.path.append(part_number)
        return update

# Update retrieval from pages.

def getUpdatesFromPage(page, request):

    """
    Get updates from the given 'page' using the 'request'. A list of update
    objects is returned.
    """

    updates = []

    # NOTE: Use the updated datetime from the page for updates.
    # NOTE: The published and updated details would need to be deduced from
    # NOTE: the page history instead of being taken from the page as a whole.

    metadata = getMetadata(page)
    updated = getUpdatedTime(metadata)

    # Get the fragment regions for the page.

    for n, (format, attributes, body) in enumerate(getFragments(page.get_raw_body())):

        update = Update()

        # Produce a fragment identifier.
        # NOTE: Choose a more robust identifier where none is explicitly given.

        update.fragment = attributes.get("fragment", str(n))
        update.title = attributes.get("summary", "Update #%d" % n)

        # Get the preferred content types available for the fragment.

        update.preferred = getPreferredOutputTypes(request, getOutputTypes(request, format))

        # Try and obtain some suitable content for the entry.
        # NOTE: Could potentially get a summary for the fragment.

        update.content = None

        if "text/html" in update.preferred:
            parser_cls = getParserClass(request, format)

            if format == "html":
                update.content = body
            elif hasattr(parser_cls, "formatForOutputType"):
                update.content = formatTextForOutputType(body, request, parser_cls, "text/html")
            else:
                fmt = request.html_formatter
                fmt.setPage(page)
                update.content = formatText(body, request, fmt, parser_cls)

            update.content_type = "text/html"

        update.page = page

        # NOTE: The anchor would be supported in the page, but this requires
        # NOTE: formatter modifications for the regions providing updates.

        update.link = page.url(request, anchor=update.fragment)
        update.updated = updated

        updates.append(update)

    return updates

# Update retrieval from message stores.

def getUpdatesFromStore(page, request):

    """
    Get updates from the message store associated with the given 'page' using
    the 'request'. A list of update objects is returned.
    """

    updates = []

    metadata = getMetadata(page)
    updated = getUpdatedTime(metadata)

    store = ItemStore(page, "messages", "message-locks")

    for n, message_text in enumerate(iter(store)):
        update = getUpdateFromMessageText(message_text, n, request)
        update.page = page
        updates.append(update)

    return updates

def getUpdateFromMessageText(message_text, message_number, request):

    "Return an update for the given 'message_text' and 'message_number'."

    update = Update()
    message = Parser().parse(StringIO(message_text))

    # Produce a fragment identifier.

    update.fragment = update.updated = getDateTimeFromRFC2822(message.get("date"))
    update.title = message.get("subject", "Update #%d" % message_number)
    update.author = message.get("moin-user")

    update.message_number = message_number

    update.content, update.content_type, update.parts, actual_author = \
        getUpdateContentFromPart(message, request)

    if actual_author:
        update.author = actual_author

    return update

def getUpdateContentFromPart(part, request):

    """
    Return decoded content, the content type, any subparts, and any author
    identity in a tuple for a given 'part'.
    """

    # Determine whether the part has several representations.

    # For a single part, use it as the update content.

    if not part.is_multipart():
        content, content_type = getPartContent(part)
        return content, content_type, None, None

    # For a collection of related parts, use the first as the update content
    # and assume that the formatter will reference the other parts.

    elif part.get_content_subtype() == "related":
        main_part = part.get_payload()[0]
        content, content_type = getPartContent(main_part)
        return content, content_type, [main_part], None

    # Encrypted content cannot be meaningfully separated.

    elif part.get_content_subtype() == "encrypted":
        try:
            part, author = getDecryptedParts(part, request)
            content, content_type, parts, _author = getUpdateContentFromPart(part, request)
            return content, content_type, parts, author
        except MoinMessageError:
            return None, part.get_content_type(), part.get_payload(), None

    # Otherwise, just obtain the parts for separate display.

    else:
        return None, part.get_content_type(), part.get_payload(), None

def getDecryptedParts(part, request):

    "Decrypt the given 'part', returning the decoded content."

    homedir = get_homedir(request)
    gpg = GPG(homedir)

    # Decrypt the part.

    if is_encrypted(part):
        text = gpg.decryptMessage(part)
        part = Parser().parse(StringIO(text))

    # Extract any signature details.

    if is_signed(part):
        result = gpg.verifyMessage(part)
        if result:
            fingerprint, identity, content = result
            return content, get_username_for_fingerprint(request, fingerprint)

    return part, None

def getPartContent(part):

    "Decode the 'part', returning the decoded payload and the content type."

    charset = part.get_content_charset()
    payload = part.get_payload(decode=True)
    return (charset and unicode(payload, charset) or payload), part.get_content_type()

def getUpdateFromPart(parent, part, part_number, request):

    "Using the 'parent' update, return an update object for the given 'part'."

    update = parent.copy(part_number)
    update.content, update.content_type, update.parts = getUpdateContentFromPart(part, request)
    return update

def getUpdatesForFormatting(update, request):

    "Get a list of updates for formatting given 'update'."

    updates = []

    # Handle multipart/alternative and other non-related multiparts.

    if update.parts:
        for n, part in enumerate(update.parts):
            update_part = getUpdateFromPart(update, part, n, request)
            updates += getUpdatesForFormatting(update_part, request)
    else:
        updates.append(update)

    return updates

# Update formatting.

def getFormattedUpdate(update, request, fmt):

    """
    Return the formatted form of the given 'update' using the given 'request'
    and 'fmt'.
    """

    # NOTE: Some control over the HTML and XHTML should be exercised.

    if update.content:
        if update.content_type == "text/html" and update.message_number is not None:
            parsers = [get_make_parser(update.page, update.message_number)]
        else:
            parsers = getParsersForContentType(request.cfg, update.content_type)

        if parsers:
            for parser_cls in parsers:
                if hasattr(parser_cls, "formatForOutputType"):
                    return formatTextForOutputType(update.content, request, parser_cls, "text/html")
                else:
                    return formatText(update.content, request, fmt, parser_cls=parser_cls)
                break
        else:
            return None
    else:
        return None

def formatUpdate(update, request, fmt):

    "Format the given 'update' using the given 'request' and 'fmt'."

    result = []
    append = result.append

    updates = getUpdatesForFormatting(update, request)
    single = len(updates) == 1

    # Format some navigation tabs.

    if not single:
        append(fmt.div(on=1, css_class="moinshare-alternatives"))

        first = True

        for update_part in updates:
            append(fmt.url(1, "#%s" % update_part.unique_id()))
            append(fmt.text(update_part.content_type))
            append(fmt.url(0))

            first = False

        append(fmt.div(on=0))

    # Format the content.

    first = True

    for update_part in updates:

        # Encapsulate each alternative if many exist.

        if not single:
            css_class = first and "moinshare-default" or "moinshare-other"
            append(fmt.div(on=1, css_class="moinshare-alternative %s" % css_class, id=update_part.unique_id()))

        # Include the content.

        append(formatUpdatePart(update_part, request, fmt))

        if not single:
            append(fmt.div(on=0))

        first = False

    return "".join(result)

def formatUpdatePart(update, request, fmt):

    "Format the given 'update' using the given 'request' and 'fmt'."

    _ = request.getText

    result = []
    append = result.append

    # Encapsulate the content.

    append(fmt.div(on=1, css_class="moinshare-content"))
    text = getFormattedUpdate(update, request, fmt)
    if text:
        append(text)
    else:
        append(fmt.text(_("Update cannot be shown for content of type %s.") % update.content_type))
    append(fmt.div(on=0))

    return "".join(result)

# Source management.

def getUpdateSources(pagename, request):

    "Return the update sources from the given 'pagename' using the 'request'."

    sources = {}

    source_definitions = getWikiDict(pagename, request)

    if source_definitions:
        for name, value in source_definitions.items():
            sources[name] = getSourceParameters(value)

    return sources

def getSourceParameters(source_definition):

    "Return the parameters from the given 'source_definition' string."

    return parseDictEntry(source_definition, ("type", "location"))

# HTML parsing support.

class IncomingHTMLSanitizer(HTMLSanitizer):

    "An HTML parser that rewrites references to attachments."

    def __init__(self, out, request, page, message_number):
        HTMLSanitizer.__init__(self, out)
        self.request = request
        self.message_number = message_number
        self.page = page

    def rewrite_reference(self, ref):
        if ref.startswith("cid:"):
            part = ref[len("cid:"):]
            action_link = self.page.url(self.request, {
                "action" : "ReadMessage", "doit" : "1",
                "message" : self.message_number, "part" : part
                })
            return action_link
        else:
            return ref

    def handle_starttag(self, tag, attrs):
        new_attrs = []
        for attrname, attrvalue in attrs:
            if attrname in self.uri_attrs:
                new_attrs.append((attrname, self.rewrite_reference(attrvalue)))
            else:
                new_attrs.append((attrname, attrvalue))
        HTMLSanitizer.handle_starttag(self, tag, new_attrs)

class IncomingMarkup(Markup):

    "A special markup processor for incoming HTML."

    def sanitize(self, request, page, message_number):
        out = getwriter("utf-8")(StringIO())
        sanitizer = IncomingHTMLSanitizer(out, request, page, message_number)
        sanitizer.feed(self.stripentities(keepxmlentities=True))
        return IncomingMarkup(unicode(out.getvalue(), "utf-8"))

class IncomingHTMLParser:

    "Filters and rewrites incoming HTML content."

    def __init__(self, raw, request, **kw):
        self.raw = raw
        self.request = request
        self.message_number = None
        self.page = None

    def format(self, formatter, **kw):

        "Send the text."

        try:
            self.request.write(formatter.rawHTML(IncomingMarkup(self.raw).sanitize(self.request, self.page, self.message_number)))
        except HTMLParseError, e:
            self.request.write(formatter.sysmsg(1) +
                formatter.text(u'HTML parsing error: %s in "%s"' % (e.msg,
                                  self.raw.splitlines()[e.lineno - 1].strip())) +
                formatter.sysmsg(0))

class MakeIncomingHTMLParser:

    "A class that makes parsers configured for messages."

    def __init__(self, page, message_number):

        "Initialise with state that is used to configure instantiated parsers."

        self.message_number = message_number
        self.page = page

    def __call__(self, *args, **kw):
        parser = IncomingHTMLParser(*args, **kw)
        parser.message_number = self.message_number
        parser.page = self.page
        return parser

def get_make_parser(page, message_number):

    """
    Return a callable that will return a parser configured for the message from
    the given 'page' with the given 'message_number'.
    """

    return MakeIncomingHTMLParser(page, message_number)

# vim: tabstop=4 expandtab shiftwidth=4
